#!/usr/bin/perl -w

use strict;

my $contig_list = $ARGV[0]; # a file containing a list of all subseq files
my $breaks_dir = $ARGV[1]; # the directory containing the above
my $pB_init = $ARGV[2]; # the value of pBinit that results in the ML allocation of breakpoints, e.g. -13.82
my $mode = $ARGV[3]; # choose one of: STARRI.2.0.noEM, STARRI.2.0.EM, STARRI.correct.noEM, or STARRI.correct.EM

my $tree_file = "trees.pB_".$pB_init.".txt";
my $common_consist_file = "pars.snp.pB_".$pB_init.".txt";
my $common_incons_file = "homoplasic.snp.pB_".$pB_init.".txt";
open TREE, ">$tree_file" || die;
open CONS, ">$common_consist_file" || die;
open INCONS, ">$common_incons_file" || die;

if ($breaks_dir !~ m/\/$/) {
    $breaks_dir .= "/";
}
my $pars_dir = $breaks_dir."dnapars.".$pB_init;

# get info for contigs and breaks
my %poly_map; # map polymorphic sites to their actual genomic/contig position
my %rev_map;
my %n; # the total number of polymorphic sites in the contig
my $Nsingle = 0; # the total number of singletons across all contigs
my %seenContig;
my %physLen; # the total number of bp in the contig, including invar sites
my %subset;
my @species;
my @poly_sites;
my @dimo_sites;
my $n;
my %split;
my %poly_variants; # the number of bases segregating at a site i
my %r; # store DIMORPHIC sites
my %dimorphic; # whether the site is dimorphic
my $poly_count = 0;
my @singletons; # list of singletons sites

my %incons_tot; # inconsistencies, summed over all sites & contigs
my %splits_tot; # legitimate splits in the parsimony tree of the block
my %bipart_tot; # both
my $most_popular_split = 0;
open CONT, $contig_list || die "Can't find $contig_list file\n";
print "reading contig info...\n";
while (<CONT>) {
    chomp;
    my ($contig, $physLen, $inform, $di, $tri, $quad, $nonInf, $lk_file, $map_file, $sge_file, $concat_file, $subset_file) = split "\t";
    next if (m/^cont/);
    my %polymorphic; # store polymorphic sites for the current contig
    unless (exists $seenContig{$contig}) {
	print "contig ", $contig,"\n";
	$seenContig{$contig} = 1;
	$physLen{$contig} = $physLen;
	unless (-e $map_file) {
	    print $map_file, " does not exist!\n";
	    exit;
	}
	open MAP, "$map_file" || die "Can't find $map_file\n";
	while (<MAP>) {
	    chomp;
	    next if (m/poly/);
	    my ($poly_pos, $cont_pos, $Nvariants) = split "\t";
	    $poly_map{$contig}{$poly_pos} = $cont_pos;
#	    $rev_map{$contig}{$cont_pos} = $poly_pos;
	    $n{$contig} = $poly_pos;
	}
	close MAP;
	
	my @DNAlines;
	open DNA, $concat_file || die "Could not open fasta file for $concat_file\n";
	while (<DNA>) {
	    push @DNAlines, $_;
	}
	close DNA;
	open FILE, $subset_file || die "Can't open infile $subset_file\n";
	while (<FILE>) {
	    chomp;
	    $subset{$_} = $_;
	}
	close FILE;
	my %DNA = readFasta(@DNAlines); # makes a hash of header:sequence
	@species = sort keys %DNA;
# puts sequences in a list to check their length
	my @sequences = values %DNA;
	my $size = length $sequences[0];
	foreach (@sequences) { # checks that all are same length
	    unless ((length $_) == $size) {
		my %reversed = reverse %DNA;
		print "This sequence is not the right length: $reversed{$_}\n";
	    }
	}
	@singletons = ();
# make new fasta, only including variable sites
	for (my $i=0; $i<$size; $i+=1) {	
	    delete $r{$i};
	    delete $dimorphic{$i};
	    my %unique_bases;
	    foreach my $species (@species) {
		my $base = substr($DNA{$species},$i,1);
		$unique_bases{$base} += 1;
	    }
	    # only retain polymorphic sites with 2 segregating bases
	    my $poly_variants = 0;
	    my $singleton = 0; # whether the site contains a singleton or not
	    my $non_sing;
	    foreach my $base (keys %unique_bases) {
		if ($unique_bases{$base} == 1) { # singleton site!
		    $singleton++;
		}
		else {
		    $non_sing++;
		}
		$poly_variants++;
	    }
	    if ($non_sing >= 2) { # a potentially informative site, though it might also contain singletons
		if ($poly_variants >= 2) { # at least 2 alleles segregating
		    $poly_count++;
		    $poly_variants{$i} = $poly_variants;
		    foreach my $sp (@species) {
			my $base = substr($DNA{$sp},$i,1);
			$polymorphic{$i}{$sp} = $base;
#			print join "\t", "poly: ", $i, $sp, $contig, $base,"\n";
		    }
		    if ($poly_variants == 2) { # exactly 2 alleles segregating => dimorphic site, to be considered in pre-filtering for perfectly parsimonious stretchs
			foreach my $sp (@species) {
			    my $base = substr($DNA{$sp},$i,1);
			    $r{$i}{$sp}{$contig} = $base;
			    $dimorphic{$i}{$contig} = $i;
			}
		    }
		}
	    }
	    elsif ($singleton >= 1) { 
		push @singletons, $i;
	    }
	}
	@poly_sites = sort numeric keys %polymorphic;
	@dimo_sites = sort numeric keys %r;
	$n = (@dimo_sites+0);
	print "polymorphic sites= ", $n{$contig},"\n";	

	%split = getSplit($contig,0,$n); # contains 2 non-overlapping arrays specifying the bipartition for site j
	# find breakpoint locations STARRI.1.5.EM.50.init.pB_-1.pM_-0.001.out
	my $break_file = $breaks_dir.$mode.".".$contig.".init.pB_".$pB_init.".out";
	print "reading breakpoint file: ", $break_file,"\n";
	my %blk_start;
	my %blk_end;
	my $blk_count=0;
	my %Nmuts; # how many changes required at this site?
	my %Nincompat; # is the site incompatible?
	open BREAKFILE, "$break_file" || die "Can't find $break_file\n";
	while (<BREAKFILE>) {
	    chomp;
	    my @breaks = split;
	    @breaks = sort numeric @breaks;
	    my $last_break = 1;
            for (my $b=1; $b<(@breaks+0); $b++) {
                my $start_blk = $last_break;
		my $end_blk = ($breaks[$b]-1);

		$blk_count++;
		$blk_start{$blk_count} = $start_blk;
		$blk_end{$blk_count} = $end_blk;
                $last_break = $breaks[$b];
		my %sp_header;
		foreach my $sp (@species) {
		    my $Nchar = length($sp);
		    my $sp_header = $sp;
		    for (my $i=1; $i<=(10 - $Nchar); $i++) {
			$sp_header .= " ";
		    }
		    $sp_header{$sp} = $sp_header;
		}
		my $start = $start_blk;
		my $end = $end_blk;		
		my $subseq_file = $pars_dir."/subseq.".$contig.".".$start."_".$end.".phy";
		my $dnapenny_out = $pars_dir."/dnacomp.".$contig.".".$start."_".$end;
		my $outfile = $pars_dir."/outfile.".$contig.".".$start."_".$end;
		my $outtree = $pars_dir."/outtree.".$contig.".".$start."_".$end;
		
# extract Nmuts, Ncompat, Nincompat from pars outfile
		open PARS, "$outfile" || die "Can't find dnacomp $outfile\n";
		my $Ncompat = 0; # number of compatible sites
		my $Nmuts = 0; # total number of mutations required in the block
		my $Nincompat = 0; # total number of double-mutation events
		my $read_steps = 0;
		my $read_compat = 0;
		my $done = 0;
		my $site_count1 = 0;
		my $site_count2 = 0;
		while (<PARS>) {
		    chomp;
		    if (m/^From/) {
			$done = 1;
			$read_steps = 0;
			$read_compat = 0;
			last;
		    }
		    last if ($done == 1);
		    if (m/total number of compatible sites is[\s]+([\d\.]+)/) {
			$Ncompat = $1;
		    }
		    if (m/\*\-\-\-\-\-\-\-\-\-\-\-\-\-\-\-\-\-\-\-\-\-\-\-\-\-\-\-\-\-\-\-\-\-\-\-\-\-\-\-\-\-$/) {
			$read_steps = 1;
			$site_count1 = 0;
		    }
		    if ($read_steps == 1) {
			if (m/\|\s+([\s\d]+)$/) {
			    for (my $i=0; $i<=(length($1)); $i++) {
				my $mut =  (substr($1,$i,1));
				if ($mut =~ m/[\d]+/) {
				    $Nmuts += $mut;
				    $site_count1++;
				    $Nmuts{$blk_count}{$site_count1} = $mut;
				}
			    }
			}
		    }
		    if (m/compatibility \(Y or N\) of each site with this tree\:/) {
			$read_steps = 0;
			$site_count2 = 0;
		    }
		    if (m/\*\-\-\-\-\-\-\-\-\-\-$/) {
			$read_compat = 1;
		    }
		    if ($read_compat == 1) {
			if (m/total number of compatible sites is[\s]+([\d\.]+)/) {
			    $read_compat = 0;
			    last;
			}
			elsif (m/[\d]+[\s]+\![\s]*([YN]+)/) {
#	    print $_,"\n";
			    for (my $site=0; $site<=(length $1); $site++) {
				my $inc = 0;
				if (substr($1, $site, 1) eq 'N') {
				    $Nincompat++;
				    $site_count2++;
				    $inc = 1;
				}
				elsif(substr($1, $site, 1) eq 'Y') {
				    $site_count2++;
				    $inc = 0;
				}
				$Nincompat{$blk_count}{$site_count2} = $inc;
			    }
			}
		    }
		}
		close PARS;		
	    }
	}
	close BREAKFILE;
    
	my $rep = 0;
	my $tree= $contig;
	if ($contig =~ m/tree([\d]+)r([\d]+)/) {
	    $tree = $1;
	    $rep = $2;
	}
	elsif ($contig =~ m/tree([\d]+)/) {
	    $tree = $1;
	}
	my $break_plot_file = $contig.".break.plot.txt";
	open PLOT, ">$break_plot_file" || die "Cant find $break_plot_file\n";
	my $mutrate_file = $breaks_dir.$mode.".".$contig.".init.pB_".$pB_init.".blocks.out";	
	open MUTRATE, "$mutrate_file" || die "Can't open $mutrate_file\n";
	print "writing breakpoint locations to file: ", $break_plot_file,"\n";
	my $curr_block = 0;
	my %physStart;
	my %physEnd;
	my $last_end = 0;
	while (<MUTRATE>) {
	    chomp;
	    my ($start_block, $end_block, $poly_map_start, $poly_map_end, $L) = split "\t";
	    $curr_block++;
	    $physStart{$curr_block} = $last_end + 1;
	    $physEnd{$curr_block} = $poly_map_end;
	    $last_end = $poly_map_end;
	    next if ($curr_block >= $blk_count);
	    print PLOT join "\t", $tree, $rep, $poly_map_end, $pB_init;
	    print PLOT "\n";
	}
	close PLOT;
	close MUTRATE;
    
	# print the identities of biparts in each block
	for (my $blk_i=1; $blk_i<=$blk_count; $blk_i++) {
	    my $start_i = $blk_start{$blk_i};
	    my $end_i   = $blk_end{$blk_i};
	    my %biparts_inBlock;
	    my %poly;
	    my $tot_incons = 0;
	    my $expect_i = 0;
	    foreach my $site (sort numeric keys %{ $Nincompat{$blk_i} }) {
		$expect_i++;
		if ($Nmuts{$blk_i}{$site} > 1) {
		    $tot_incons++;
		}
	    }
	    my %splits_in_block;
	    my %incons_in_block;
	    my $i_count = 0;
	    for (my $site_i = ($start_i-1); $site_i <= ($end_i-1); $site_i++) {
		$i_count++;
		
		if (exists $dimorphic{$poly_sites[$site_i]} && exists $split{$contig}{$poly_sites[$site_i]}{1} ) {
#		    if (exists $dimorphic{$dimo_sites[$site_i]} && exists $split{$dimo_sites[$site_i]}{1} ) {
		    my @group1_1 = @{ $split{$contig}{$poly_sites[$site_i]}{1} };
		    my $group1_2 = diffs(\@species, \@group1_1);
		    my @group1_2 = @$group1_2;
		    my $groupA = "";
		    my $groupB = "";
		    foreach my $sp (sort @{ $split{$contig}{$poly_sites[$site_i]}{1} } ) {
			$groupA .= $sp.".";
		    }
		    foreach my $sp (sort @{ $split{$contig}{$poly_sites[$site_i]}{2} } ) {
			$groupB .= $sp.".";
		    }
		    my $min = $groupA;
		    my $max = $groupB;
		    if ( (@group1_1+0) > (@group1_2+0) ) {
			$min = $groupB;
			$max = $groupA;
		    }
		    if ($Nmuts{$blk_i}{$i_count} > 1) {
			$incons_in_block{$min}++;
			$incons_tot{$min}++;
		    }
		    else {
			$splits_in_block{$min}++;
			$splits_tot{$min}++;
		    }
		    $bipart_tot{$min}++;
		    if ($bipart_tot{$min} > $most_popular_split) {
			$most_popular_split = $bipart_tot{$min};
		    }
		}
	    }
	    my $tree_str = "";
	    my $Nsnp_inTree=0;
	    my $inTree_str = "";
	    foreach my $split (sort keys %splits_in_block) {
		$tree_str .= $split."_";
		$Nsnp_inTree += $splits_in_block{$split};
		$inTree_str .= $splits_in_block{$split}."_";
	    }
	    my $incons_str = "";
	    my $Nsnp_homoplasic=0;
	    my $homoplasic_str = "";
	    foreach my $split (sort keys %incons_in_block) {
		$incons_str .= $split."_";
		$Nsnp_homoplasic += $incons_in_block{$split};
		$homoplasic_str .= $incons_in_block{$split}."_";
	    }
	    if ($tree_str eq "") {
		$tree_str = 'na';
	    }
	    if ($incons_str eq "") {
		$incons_str = 'na';
	    }
	    if ($inTree_str eq "") {
		$inTree_str = 'na';
	    }
	    if ($homoplasic_str eq "") {
		$homoplasic_str = 'na';
	    }
	    print TREE join "\t", $contig, $blk_i, $physStart{$blk_i}, $physEnd{$blk_i}, $Nsnp_inTree, $Nsnp_homoplasic, $tree_str, $inTree_str, $incons_str, $homoplasic_str;
	    print TREE "\n";
	}
    }
}	    
close CONT;

print "common biparts in pars trees\n";
for (my $i=$most_popular_split; $i>0; $i-=1) {
    foreach my $split (sort keys %splits_tot) {
	if ($splits_tot{$split} == $i) {
	    print CONS join "\t", $i, $split;
	    print CONS "\n";
	}
    }
}

print "common homoplasies\n";
for (my $i=$most_popular_split; $i>0; $i-=1) {
    foreach my $split (sort keys %incons_tot) {
	if ($incons_tot{$split} == $i) {
	    print INCONS join "\t", $i, $split;
	    print INCONS "\n";
	}
    }
}

#returns a hash of sequence indexed by fasta headers
sub readFasta{
    my %fasta;
    my $line;
    my $length=0;
    my $header;

    for $line (@_){
        chomp $line;

        #find header lines
	if ($line =~ /^>(.+)/) {
            $header = $1;
#	    print $header,"\n";
            next;
        }

        #for non-header lines...
        if(defined($header)){
	    # only include species in the subset of interest
	    next unless (exists $subset{$header});
            #get rid of spaces
            $line =~ s/\s+//g;
            #append to the current sequence stored in %fasta
            $fasta{$header} .= $line;
        }
    }
    #return sequences indexed by headers
    return %fasta;
}

sub numeric {
    $a <=> $b;
}

sub getSplit { # convert a site to 2 non-overlapping arrays
    my ($contig,$i,$j) = @_; # the sites of interest
    my %group;
#    print "sub_getSplit::\n";
    for (my $pos=$i; $pos< $j; $pos+=1) {
#	print "position= ", $pos,"\n";
	my $map = $dimo_sites[$pos];
	if (exists $r{$map}{$species[1]}{$contig}) { # Only include dimorphic sites
	    my $ref_base = $r{$map}{$species[1]}{$contig};

	    foreach my $sp (@species) {
		if ($r{$map}{$sp}{$contig} eq $ref_base) {
		    push @{ $group{$contig}{$map}{1} }, $sp;
#		    print "add ", $sp, " to group1\n";
		}
		else {
		    push @{ $group{$contig}{$map}{2} }, $sp;
#		    print "add ", $sp, " to group2\n";
		}
	    }
	    # group1 always contains the shorter list of strain names
	    my @min;
	    my @max;
	    if ( (@{ $group{$contig}{$map}{1} }+0) < (@{ $group{$contig}{$map}{2} }+0) ) {
		@min = sort @{ $group{$contig}{$map}{1} };
		@max = sort @{ $group{$contig}{$map}{2} };
	    }
	    else {
		@min = sort @{ $group{$contig}{$map}{2} };
		@max = sort @{ $group{$contig}{$map}{1} };
	    }
	    @{ $group{$contig}{$map}{1} } = @min;
	    @{ $group{$contig}{$map}{2} } = @max;
#	    print join "\t", "@{ $group{$map}{1} }", "@{ $group{$map}{2} }";
#	    print "\n";
	}
	else {
#	    print join "\t", "no r: ", $map, $dimo_sites[$pos], $pos, $contig,"\n";
	}
    }
    return %group;
}

sub diffs { # returns differences between 2 arrays
    my ($array1, $array2) = @_;
    my @array1 = @$array1;
    my @array2 = @$array2;
    my @union = my @intersection = my @difference = ();
    my %count = ();
    foreach my $element (@array1, @array2) {
	$count{$element}++;
    }
    foreach my $element (keys %count) {
	push @union, $element;
	push @{ $count{$element} > 1 ? \@intersection : \@difference }, $element;
    }
    return \@difference;
}

sub intersect { # returns the intersection between 2 arrays                                                                                                                                                     
    my ($array1, $array2) = @_;
    my @array1 = @$array1;
    my @array2 = @$array2;
    my @union = my @intersection = my @difference = ();
    my %count = ();
    foreach my $element (@array1, @array2) {
        $count{$element}++;
    }
    foreach my $element (keys %count) {
        push @union, $element;
        push @{ $count{$element} > 1 ? \@intersection : \@difference }, $element;
    }
    return \@intersection;
}
