package edu.mit.polzlab.polymorphisms;

import java.util.*;

import cern.colt.list.DoubleArrayList;
import cern.colt.list.IntArrayList;
import cern.jet.random.ChiSquare;
import cern.jet.random.engine.MersenneTwister;

import edu.mit.polzlab.polymorphisms.util.PmorphStats;
import edu.mit.polzlab.polymorphisms.util.PmorphUtils;

public class TopoBlocks {
	
	public HashMap [] disscordantPhyloInfIndex(HashMap aln, int window, int step, int thresh){

		HashMap allp = PmorphUtils.allPolymorphisms(aln);
		HashMap pmcounts = PmorphStats.splitFreqs(aln);
		//here should come some more informed threshold
		
		for(Iterator it = pmcounts.keySet().iterator(); it.hasNext();){
			String id = (String)it.next();
			if(((Integer)pmcounts.get(id)) < thresh || id.split(";").length == 1){ //removes singletons!
				allp.remove(id);
			}
		}
		
		HashMap [] pruned = PmorphUtils.removeNonPolymSites(aln, allp);
		HashMap newseq = pruned[0];
		HashMap posmap = pruned[1];
		
		int N = ((String)newseq.get(newseq.keySet().iterator().next())).length();
		
		Object [] data = PmorphUtils.getCompatibilityMatrix(allp);
		List allsplitids = (List)data[0];
		int [][] compatibilities = (int [][])data[1];
		
		HashMap dpiis = new HashMap();
		for(int i = 0; i < N-window; i=i+step){
			double dpii = blockDisscordantPhylInfIndex(newseq, i, i+window, allsplitids, compatibilities);
			for(int j = i; j < i+window; j++){
				DoubleArrayList d = (DoubleArrayList)dpiis.get(j);
				if(d == null){
					d = new DoubleArrayList();
					dpiis.put(j,d);
				}
				d.add(dpii);
			}
		}
		for(int i = N-1; i >= N-window; i=i-step){
			double dpii = blockDisscordantPhylInfIndex(newseq, i-window, i, allsplitids, compatibilities);
			for(int j = i-window; j < i; j++){
				DoubleArrayList d = (DoubleArrayList)dpiis.get(j);
				if(d == null){
					d = new DoubleArrayList();
					dpiis.put(j,d);
				}
				d.add(dpii);
			}
		}
		
		HashMap [] ret = new HashMap[2];
		ret[0] = dpiis;
		ret[1] = posmap;
		return ret;
	}
	
	
	/** test for phylogenetic anormalities in alignment with chi-square test
	 * 
	 * @param aln
	 * @param window
	 * @param step
	 * @param thresh
	 * @return
	 */
	public HashMap [] chiSquareBreakPointIndex(HashMap aln, int window, int step, int thresh){
		
		HashMap allp = PmorphUtils.allPolymorphisms(aln);
		HashMap pmcounts = PmorphStats.splitFreqs(aln);
		//here should come some more informed threshold
		
		for(Iterator it = pmcounts.keySet().iterator(); it.hasNext();){
			String id = (String)it.next();
			if(((Integer)pmcounts.get(id)) < thresh || id.split(";").length == 1){ //removes singletons!
				allp.remove(id);
			}
		}
		
		HashMap [] pruned = PmorphUtils.removeNonPolymSites(aln, allp);
		HashMap newseq = pruned[0];
		HashMap posmap = pruned[1];
		
		Object [] data = PmorphUtils.getCompatibilityMatrix(allp);
		
		int N = ((String)newseq.get(newseq.keySet().iterator().next())).length();
		
		HashMap chis = new HashMap();
		for(int i = 0; i < N-window; i=i+step){
			double chi = blockChiSquareTest(newseq, i, i+window);
			for(int j = i; j < i+window; j++){
				DoubleArrayList d = (DoubleArrayList)chis.get(j);
				if(d == null){
					d = new DoubleArrayList();
					chis.put(j,d);
				}
				d.add(chi);
			}
		}
		for(int i = N-1; i >= N-window; i=i-step){
			double chi = blockChiSquareTest(newseq, i-window, i);
			for(int j = i-window; j < i; j++){
				DoubleArrayList d = (DoubleArrayList)chis.get(j);
				if(d == null){
					d = new DoubleArrayList();
					chis.put(j,d);
				}
				d.add(chi);
			}
		}
		
		HashMap [] ret = new HashMap[3];
		ret[0] = chis;
		ret[1] = posmap;
		ret[2] = newseq;
		return ret;
	}
	
	/** identify recomb blocks (low information)
	 * 
	 * @param aln
	 * @param window
	 * @param step
	 * @return
	 */
	public HashMap [] chiSquareRecombIndex(HashMap aln, int window, int step, HashMap globalsplitf, int totN){
		
		HashMap allp = PmorphUtils.allPolymorphisms(aln);
		HashMap pmcounts = PmorphStats.splitFreqs(aln);
		//here should come some more informed threshold
		
		for(Iterator it = pmcounts.keySet().iterator(); it.hasNext();){
			String id = (String)it.next();
			if(id.split(";").length == 1){ //removes singletons!
				allp.remove(id);
			}
		}
		
		HashMap [] pruned = PmorphUtils.removeNonPolymSites(aln, allp);
		HashMap newseq = pruned[0];
		HashMap posmap = pruned[1];
				
		int N = ((String)newseq.get(newseq.keySet().iterator().next())).length();
		
		HashMap chis = new HashMap();
		for(int i = 0; i < N-window; i=i+step){
			double chi = chiSquareTest(newseq, i, i+window, globalsplitf, totN);
			for(int j = i; j < i+window; j++){
				DoubleArrayList d = (DoubleArrayList)chis.get(j);
				if(d == null){
					d = new DoubleArrayList();
					chis.put(j,d);
				}
				d.add(chi);
			}
		}
		for(int i = N-1; i >= N-window; i=i-step){
			double chi = chiSquareTest(newseq, i-window, i, globalsplitf, totN);
			for(int j = i-window; j < i; j++){
				DoubleArrayList d = (DoubleArrayList)chis.get(j);
				if(d == null){
					d = new DoubleArrayList();
					chis.put(j,d);
				}
				d.add(chi);
			}
		}
		
		HashMap [] ret = new HashMap[2];
		ret[0] = chis;
		ret[1] = posmap;
		return ret;
	}
	
	
	/** returns a map with all observed topologies and their prob per windows 
	 * 
	 * @param aln
	 * @param window
	 * @param step
	 * @param thresh
	 * @return
	 */
	public HashMap topologySpectrum(HashMap aln, int window, int step, int thresh){
		HashMap topoprobs = new HashMap();
		int N = ((String)aln.get(aln.keySet().iterator().next())).length();
		
		HashMap allp = PmorphUtils.allPolymorphisms(aln);
		HashMap pmcounts = PmorphStats.splitFreqs(aln);
		//here should come some more informed threshold
		
		for(Iterator it = pmcounts.keySet().iterator(); it.hasNext();){
			String id = (String)it.next();
			if(((Integer)pmcounts.get(id)) < thresh){
				allp.remove(id);
			}
		}
		
		Object [] data = PmorphUtils.getCompatibilityMatrix(allp);
		List allsplitids = (List)data[0];
		int [][] compatibilities = (int [][])data[1];
		
		int count = 0;
		for(int i = 0; i < N-window; i=i+step){
			HashMap block_topoprobs = blockTopologies(aln, i, i+window, allsplitids, compatibilities);
			for(Iterator it = topoprobs.keySet().iterator(); it.hasNext();){
				String id = (String)it.next();
				DoubleArrayList d = (DoubleArrayList)topoprobs.get(id);
				if(block_topoprobs.get(id) == null)
					d.add(0);
			}
			for(Iterator it = block_topoprobs.keySet().iterator(); it.hasNext();){
				String id = (String)it.next();
				DoubleArrayList d = (DoubleArrayList)topoprobs.get(id);
				if(d == null){
					d = new DoubleArrayList();
					for(int x =0; x < count; x++){
						d.add(0);
					}
					topoprobs.put(id, d);
				}
				d.add((Double)block_topoprobs.get(id));
			}
			count++;
		}
		
		return topoprobs;
	}
	
	/**
	 * 
	 * @param aln
	 * @param pos0
	 * @param pos1
	 * @param allsplitids
	 * @param compatibilities
	 * @return double - an index indicating the average overlap between all pairs of observed topologies, considering bipartition weights 
	 */
	public double blockDisscordantPhylInfIndex(HashMap aln, int pos0, int pos1, List allsplitids, int [][] compatibilities){
		int [] scounts = new int[allsplitids.size()];
		ArrayList topologies = new ArrayList();
		for(int i = pos0; i < pos1; i++){
			List split = PmorphUtils.getPolymorphismSplit(aln, i);
			if(split == null)
				continue;
			updateTopologies(split, allsplitids, scounts, compatibilities, topologies);
		}
		
		double totalpm = 0;
		for(int i = 0; i < scounts.length; i++){
			totalpm += scounts[i];
		}
		
		double dpii = 0;
		double count = 0;
		for(int i = 0; i < topologies.size()-1; i++){
			IntArrayList topoi = (IntArrayList)topologies.get(i);
			for(int j = i+1; j < topologies.size(); j++){
				IntArrayList topoj = (IntArrayList)topologies.get(j);
				int overlap = 0;
				int noverlapi = 0;
				int noverlapj = 0;
				for(int x = 0; x < topoi.size(); x++){
					//comparing i -> j direction
					int splitidx = topoi.get(x);
					if(topoj.contains(splitidx)) //overlap
						overlap += scounts[splitidx];
					else noverlapi += scounts[splitidx];	
				}
				for(int x = 0; x < topoj.size(); x++){
					//comparing j -> j direction
					int splitidx = topoj.get(x);
					if(!topoi.contains(splitidx)) //overlap
						noverlapj += scounts[splitidx];
				}
				dpii += (double)Math.min(noverlapi,noverlapj)/(overlap+noverlapi+noverlapj);
				count++;
			}
		}
		if(count == 0)
			return 0;
		return dpii/count;
	}
	
	
	/**
	 * Fisher extact test for the 50-50 partition of a window
	 *
	 */
	public double blockChiSquareTest(HashMap aln, int pos0, int pos1){
		int middpoint = Math.round((pos0+pos1)/2);
		HashMap splitf1 = PmorphStats.splitFreqs(aln,pos0,middpoint-1);
		HashMap splitf2 = PmorphStats.splitFreqs(aln,middpoint,pos1);
		
		HashMap rows = new HashMap();
		HashMap rowsums = new HashMap();
		int sumcol1 = 0;
		int sumcol2 = 0;
		double N = 0;
		for(Iterator it = splitf1.keySet().iterator(); it.hasNext();){
			String key = (String)it.next();
			int [] counts = (int [])rows.get(key);
			if(counts == null){
				counts = new int[2];
			}
			counts[0] = (Integer)splitf1.get(key);
			sumcol1 += counts[0];
			rows.put(key, counts);
			rowsums.put(key, counts[0]+counts[1]);
		}
		for(Iterator it = splitf2.keySet().iterator(); it.hasNext();){
			String key = (String)it.next();
			int [] counts = (int [])rows.get(key);
			if(counts == null){
				counts = new int[2];
			}
			counts[1] = (Integer)splitf2.get(key);
			sumcol2 += counts[1];
			rows.put(key, counts);
			rowsums.put(key, counts[0]+counts[1]);
		}
		
		if(rowsums.size()==1)
			return 1;
		
		N = sumcol1 + sumcol2;
		double chis = 0;
		
		//williams' correction
		double n = pos1-pos0;
		double qrow = 0;
				
		for(Iterator it = rowsums.keySet().iterator(); it.hasNext();){
			
			String key = (String)it.next();
			int row = (Integer)rowsums.get(key);
			double ei = row*sumcol1/N;
			double ej = row*sumcol2/N;
			
			int oi = ((int [])rows.get(key))[0];
			int oj = ((int [])rows.get(key))[1];
			
			//chis += Math.pow(Math.abs(oi - ei)-0.5,2)/ei + Math.pow(Math.abs(oj - ej)-0.5,2)/ej; //Yates
			chis += Math.pow(oi - ei,2)/ei + Math.pow(oj - ej,2)/ej;
			
			//for williams correction
			qrow += 1/row;
		}
		
		double qcol = 1/sumcol1+1/sumcol2;
		double q = 1+(n*qrow - 1)*(n*qcol - 1)/(6*n*(rowsums.size() - 1));
		
		ChiSquare csq = new ChiSquare((rowsums.size()-1), new MersenneTwister((int)System.currentTimeMillis()));
		
		double cum = csq.cdf(chis); // /q
		return (1-cum);
		//return chis;
		
	}
	
	/** block chisquare test adding pseudocount rows with laplacian rule */
	
	public double blockChiSquareTestPsCount(HashMap aln, int pos0, int pos1, int psrows){
		int middpoint = Math.round((pos0+pos1)/2);
		HashMap splitf1 = PmorphStats.splitFreqs(aln,pos0,middpoint-1);
		HashMap splitf2 = PmorphStats.splitFreqs(aln,middpoint,pos1);
		
		HashMap rows = new HashMap();
		HashMap rowsums = new HashMap();
		int sumcol1 = psrows;
		int sumcol2 = psrows;
		double N = 0;
		for(Iterator it = splitf1.keySet().iterator(); it.hasNext();){
			String key = (String)it.next();
			int [] counts = (int [])rows.get(key);
			if(counts == null){
				counts = new int[2];
			}
			counts[0] = (Integer)splitf1.get(key);
			sumcol1 += counts[0];
			rows.put(key, counts);
			rowsums.put(key, counts[0]+counts[1]);
		}
		for(Iterator it = splitf2.keySet().iterator(); it.hasNext();){
			String key = (String)it.next();
			int [] counts = (int [])rows.get(key);
			if(counts == null){
				counts = new int[2];
			}
			counts[1] = (Integer)splitf2.get(key);
			sumcol2 += counts[1];
			rows.put(key, counts);
			rowsums.put(key, counts[0]+counts[1]);
		}
		
		if(rowsums.size()==1)
			return 1;
		
		N = sumcol1 + sumcol2 + rowsums.size()*2 + psrows*2;
		double chis = 0;
				
		for(Iterator it = rowsums.keySet().iterator(); it.hasNext();){
			
			String key = (String)it.next();
			int row = (Integer)rowsums.get(key);
			double ei = (row+2)*(sumcol1+rowsums.size()+psrows)/N;
			double ej = (row+2)*(sumcol2+rowsums.size()+psrows)/N;
			
			int oi = ((int [])rows.get(key))[0]+1;
			int oj = ((int [])rows.get(key))[1]+1;
			
			chis += Math.pow(oi - ei,2)/ei + Math.pow(oj - ej,2)/ej;
		}
		
		//pseudocount rows
		for(int i = 0; i < psrows; i++){
			
			double ei = 2*(sumcol1+rowsums.size()+psrows)/N;
			double ej = 2*(sumcol2+rowsums.size()+psrows)/N;
			
			int oi = 1;
			int oj = 1;
			
			chis += Math.pow(oi - ei,2)/ei + Math.pow(oj - ej,2)/ej;
		}
		
		ChiSquare csq = new ChiSquare((rowsums.size()+psrows-1), new MersenneTwister((int)System.currentTimeMillis()));
		
		double cum = csq.cdf(chis); // /q
		return (1-cum);
		//return chis;
		
	}
	
	/** calculates chi-square statistic of whether block represents background bipartition dist.
	 * 
	 * @param aln
	 * @param pos0
	 * @param pos1
	 * @return
	 */
	public double chiSquareTest(HashMap aln, int pos0, int pos1, HashMap globalcounts, int totN){
		
		HashMap splitf1 = PmorphStats.splitFreqs(aln,pos0,pos1);
		if(splitf1.size() == 1)
			return 0;
		double chis = 0;
		int window = pos1-pos0;
		
		for(Iterator it = globalcounts.keySet().iterator(); it.hasNext();){
			String key = (String)it.next();
			int count = (Integer)globalcounts.get(key);
			double e = (double)window*count/totN;
			Integer o = (Integer)splitf1.get(key);
			if(o == null)
				o = 0;
			chis = Math.pow(o-e,2)/e;
		}
		ChiSquare csq = new ChiSquare((splitf1.size()-1), new MersenneTwister((int)System.currentTimeMillis()));
		double cum = csq.cdf(chis);
		return 1-cum;
		//return chis;
		
	}
	
	
	/**
	 * 
	 * @param aln
	 * @param pos0
	 * @param pos1
	 * @param allsplitids
	 * @param compatibilities
	 * @return a map with the topologies observed in a block and their probabilities 
	 */
	public HashMap blockTopologies(HashMap aln, int pos0, int pos1, List allsplitids, int [][] compatibilities){
		int [] scounts = new int[allsplitids.size()];
		ArrayList topologies = new ArrayList();
		for(int i = pos0; i < pos1; i++){
			List split = PmorphUtils.getPolymorphismSplit(aln, i);
			if(split == null || split.size() == 1) //only di-morphic, no singletons
				continue;
			updateTopologies(split, allsplitids, scounts, compatibilities, topologies);
		}
		
		double totalpm = 0;
		for(int i = 0; i < scounts.length; i++){
			totalpm += scounts[i];
		}
		
		HashMap topomap = new HashMap();
		for(int i = 0; i < topologies.size(); i++){
			IntArrayList topo = (IntArrayList)topologies.get(i);
			topo.sort();
			String topoid = String.valueOf(topo.get(0));
			double sumtopo = 0;
			for(int j = 1; j < topo.size(); j++){
				int splitidx = topo.get(j);
				topoid += ";"+splitidx;
				sumtopo += scounts[splitidx];
			}
			topomap.put(topoid, sumtopo/totalpm);
		}
		return topomap;
	}
	
	
	/**
	 * 
	 * @param split
	 * @param allsplitids
	 * @param scounts
	 * @param compatibilities
	 * @param topologies
	 */
	public void updateTopologies(List split, List allsplitids, int [] scounts, int [][] compatibilities, List topologies){
		String splitid = PmorphUtils.getSplitId(split);
		int idx = allsplitids.indexOf(splitid);
		if(idx == -1)
			return;
		scounts[idx]++;
		if(scounts[idx] > 1)
			return;
		
		if(topologies.isEmpty()){
			IntArrayList topo = new IntArrayList();
			topo.add(idx);
			topologies.add(topo);
			return;
		}else{
			boolean isnewtopo = true;
			IntArrayList newtopo = new IntArrayList();
			newtopo.add(idx);
			for(int i = 0; i < topologies.size(); i++){
				boolean belongs = true;
				IntArrayList topo = (IntArrayList)topologies.get(i);
				for(int j = 0; j < topo.size(); j++){
					int idxj = topo.get(j);
					if(compatibilities[idx][idxj] == 0){
						belongs = false;
					}else if(isnewtopo){
						if(!newtopo.contains(idxj))
							newtopo.add(idxj);
					}
				}
				if(belongs){
					topo.add(idx);
					isnewtopo = false;
				}
			}
			if(isnewtopo)
				topologies.add(newtopo);
		}
	}
}
