package edu.mit.polzlab.polymorphisms;

import java.util.HashMap;
import java.util.Iterator;

import cern.colt.list.DoubleArrayList;
import cern.colt.list.IntArrayList;
import cern.jet.stat.Descriptive;

import uu.binf.utils.HashMapUtils;
import edu.mit.polzlab.polymorphisms.util.PmorphUtils;

public class TopoBlocksMain {
	
	public static void main(String [] args) throws Exception{
		
		if(args.length < 1){
			System.out.println("--------------------------");
			System.out.println("Analysis of tree topologies in real genome / contig:");
			System.out.println("Usage: runtopoblocks.sh [full-path-to-contig-alignment] [window-size, default = 150]");
			System.out.println("--------------------------");
			System.exit(1);
		}
		
		String contig = args[0];
		int window = 150;
		if(args.length > 1)
			window = Integer.parseInt(args[1]);
		
		HashMap seq = HashMapUtils.fastaToHashMap(contig);
			
		/*seq.remove("12B01");
		seq.remove("12F01");
		seq.remove("13B01");
		seq.remove("ZS63");
		seq.remove("S301");*/
				
		TopoBlocks tb = new TopoBlocks(); 
		
		int step = 1;
		
		HashMap [] datat = tb.disscordantPhyloInfIndex(seq, window, step, 1);
		HashMap [] data = tb.chiSquareBreakPointIndex(seq, window, step, 1);
		
		HashMap scores = datat[0];
		HashMap chi = data[0];
		HashMap posmap = data[1];
		HashMap newseq = data[2];
		HashMap binseq = PmorphUtils.getBinarySNPs(newseq);
		
		System.out.print("pos posrea mean_t sd_t mean_chi signf");
		for(Iterator it = binseq.keySet().iterator(); it.hasNext();)
			System.out.print(" " + it.next());
		System.out.println("");
		
		// * 1-02
		// ** 1-03
		// *** 1-04
		// **** 1-05
		// ***** 1-06
		
		for(int id = 0; id < scores.size(); id++){
			if(!scores.containsKey(id))
				continue;
			DoubleArrayList dt = (DoubleArrayList)scores.get(id);
			double meant = Descriptive.mean(dt);
			double sdt = Math.sqrt(Descriptive.sampleVariance(dt, meant));
			if(!(sdt < 0) && !(sdt > 0)) //NA test
				sdt = 0;
			
			DoubleArrayList dc = (DoubleArrayList)chi.get(id);
			double meanc = Descriptive.mean(dc);
			
			String mark = "-";
			if(meant > 0.1){
				if(meanc < 0.000001)
					mark = "*****";
				else if(meanc < 0.00001)
					mark = "****";
				else if(meanc < 0.0001)
					mark = "***";
				else if(meanc < 0.001)
					mark = "**";
				else if(meanc < 0.01)
					mark = "*";
			}
			
			StringBuffer buff = new StringBuffer();
			for(Iterator it0 = binseq.keySet().iterator(); it0.hasNext();){
				String key = (String)it0.next();
				buff.append(" "+((IntArrayList)binseq.get(key)).get(id));
			}
			System.out.println(id + " " + posmap.get(id) + " " + meant + " " + sdt + " " + meanc + " " + mark + buff.toString());
		}
		
	}
	
}
