package edu.mit.polzlab.polymorphisms.util;

import java.io.BufferedWriter;
import java.io.FileWriter;
import java.util.*;

import uu.binf.utils.HashMapUtils;

import cern.colt.list.IntArrayList;

public class PmorphStats {
		
	public static HashMap splitFreqs(HashMap seq, int pos0, int pos1){
		
		ArrayList splits = new ArrayList();
		for(int i = pos0; i < pos1; i++){
			List split = PmorphUtils.getPolymorphismSplit(seq, i);
			if(split == null)
				continue;
			Collections.sort(split);
			String splitid = (String)split.get(0);
			for(int j = 1; j < split.size(); j++){
				splitid += ";"+split.get(j);
			}
			splits.add(splitid);
		}
		HashMap splitcounts = HashMapUtils.listToCountMap(splits);
		return splitcounts;
	}
	
	public static HashMap splitFreqs(HashMap seq){
		
		int N = ((String)seq.get(seq.keySet().iterator().next())).split("").length;
		return splitFreqs(seq,0,N-1);
	
	}
	
	public static void savePmorphCompatibilityNet(HashMap splitcounts,String filename) throws Exception{
		
		BufferedWriter writer = new BufferedWriter(new FileWriter(filename));
		writer.write("graph G {\n");
		writer.write("size=\"10,15\";\noverlap=scale;\nnode [style=filled];\n");
		
		ArrayList splits = new ArrayList();
		int count = 0;
		for(Iterator it = splitcounts.keySet().iterator(); it.hasNext();){
			String id = (String)it.next();
			if(((Integer)splitcounts.get(id)) < 80)
				continue;
			writer.write("\"" + id + "\" [fontsize=\""+20*Math.log(((Integer)splitcounts.get(id)))+"\"," +
					"width=\""+5*Math.log(((Integer)splitcounts.get(id)))+"\",height=\""+5*Math.log(((Integer)splitcounts.get(id)))+"\"];\n");
			splits.add(id);
			count++;
		}
		for(int i = 0; i < splits.size()-1; i++){
			String id1 = (String)splits.get(i);
			List idl1 = Arrays.asList(id1.split(";"));
			for(int j = i+1; j < splits.size(); j++){
				String id2 = (String)splits.get(j);
				List idl2 = Arrays.asList((id2).split(";"));
				if(!PmorphUtils.compatible(idl1, idl2)){
					writer.write("\"" + id1 + "\" -- \""
							+ id2 + "\" [color=\"black\"];\n");
				}
			}
		}
		writer.write("}");
		writer.flush();
		writer.close();
	}
	
	public static IntArrayList splitSizeDist(HashMap seq){
		
		IntArrayList sizes = new IntArrayList();
		int N = ((String)seq.get(seq.keySet().iterator().next())).split("").length;
		for(int i = 0; i < N-1; i++){
			int s = PmorphUtils.getSplitSize(seq, i);
			if(s >= 0)
				sizes.add(s);
		}
		return sizes;
	}
	
	
	public static HashMap tipPmorphPerTaxa(HashMap seq){
		int ntaxa = seq.size();
		HashMap counts = new HashMap();
		int N = ((String)seq.get(seq.keySet().iterator().next())).split("").length;
		for(int i = 0; i < N-1; i++){
			List split = PmorphUtils.getPolymorphismSplit(seq, i);
			if(split == null)
				continue;
			String sid = PmorphUtils.getSplitId(split);
			if(sid.split(";").length == 1){
				Integer c = (Integer)counts.get(sid);
				if(c == null)
					counts.put(sid, 1);
				else
					counts.put(sid,c+1);
			}
		}
		return counts;
	}
	
	
}
