package edu.mit.polzlab.polymorphisms.util;

import java.io.*;
import java.util.*;

import cern.colt.list.IntArrayList;
import cern.jet.random.engine.MersenneTwister;

import uu.binf.utils.ListUtils;
import uu.binf.utils.MatrixUtils;

public class PmorphUtils {
	
	/** 
	 * all observed splits in the data sets (split divides alignment column in two)
	 * */
	public static HashMap allPolymorphisms(HashMap aln){
		
		int N = ((String)aln.get(aln.keySet().iterator().next())).length();
		HashMap polys = new HashMap();
		for(int i = 0; i < N; i++){
			List split = getPolymorphismSplit(aln, i);
			if(split == null)
				continue;
			String splitid = getSplitId(split);
			polys.put(splitid, split); //redundant, just to save some time later.
		}
		return polys;		
	}
	
	/** 
	 * precomputed compatibilites between all observed splits 
	 * */
	public static Object[] getCompatibilityMatrix(HashMap polys){
		
		int [][] matrix = new int[polys.size()][polys.size()];
		ArrayList keys = new ArrayList(polys.keySet());
		for(int i = 0; i < keys.size()-1; i++){
			String idi = (String)keys.get(i);
			List spliti = (List)polys.get(idi);
			for(int j = i+1; j < keys.size(); j++){
				String idj = (String)keys.get(j);
				List splitj = (List)polys.get(idj);
				if(compatible(spliti, splitj)){
					matrix[i][j] = 1;
					matrix[j][i] = 1;
				}
			}
		}
		Object [] ret = new Object[2];
		ret[0] = keys;
		ret[1] = matrix;
		return ret;
	}
	
	
	public static int getSplitSize(HashMap aln, int pos){
		String obsnt = "";
		int [] ntsplits = new int[4]; 
		for(Iterator it = aln.keySet().iterator(); it.hasNext();){
			String id = (String)it.next();
			String seq = (String)aln.get(id);
			char nt = seq.charAt(pos);
			if(obsnt.indexOf(nt)==-1){
				obsnt += nt;
				if(obsnt.length() > 2)
					return -1;
			}
			if(nt == 'A')
				ntsplits[0]++;
			else if(nt == 'C')
				ntsplits[1]++;
			else if(nt == 'G')
				ntsplits[2]++;
			else if(nt == 'T')
				ntsplits[3]++;			
		}
		if(obsnt.length() == 1)
			return 0;
		
		int size = 0;
		for(int i = 0; i < 4; i++){
			if(ntsplits[i] > size && ntsplits[i] > 0)
				size = ntsplits[i];
		}
		
		return aln.size()-size;
		
	}

	
	public static List getPolymorphismSplit(HashMap aln, int pos){
		String obsnt = "";
		ArrayList [] labels = {new ArrayList(),new ArrayList(),new ArrayList(),new ArrayList()}; 
		for(Iterator it = aln.keySet().iterator(); it.hasNext();){
			String id = (String)it.next();
			String seq = (String)aln.get(id);
			char nt = seq.charAt(pos);
			if(obsnt.indexOf(nt)==-1){
				obsnt += nt;
				/** TODO
				 * checking that snps are monomorphic
				 *  this condition needs to be relaxed for when more species are available */
				if(obsnt.length() > 2)
					return null;
			}
			if(nt == 'A')
				labels[0].add(id);
			else if(nt == 'C')
				labels[1].add(id);
			else if(nt == 'G')
				labels[2].add(id);
			else if(nt == 'T')
				labels[3].add(id);
			else return null;
		}
		if(obsnt.length() == 1)
			return null;
		
		int idx = -1;
		int min = aln.size();
		
		if(labels[0].size() > 0 && labels[0].size() < min){
			min = labels[0].size();
			idx = 0;
		}
		if(labels[1].size() > 0 && labels[1].size() < min){
			min = labels[1].size();
			idx = 1;
		}
		if(labels[2].size() > 0 && labels[2].size() < min){
			min = labels[2].size();
			idx = 2;
		}
		if(labels[3].size() > 0 && labels[3].size() < min){
			idx = 3;
		}
		
		return labels[idx];
	}
	
	public static boolean compatible(List split1, List split2){
		int inters = ListUtils.sizeIntersection(split1, split2);
		if(inters < Math.min(split1.size(),split2.size()) && inters > 0){
			return false;
		}
		return true;
	}
	
	public static String getSplitId(List split){
		Collections.sort(split);
		String splitid = (String)split.get(0);
		for(int j = 1; j < split.size(); j++){
			splitid += ";"+split.get(j);
		}
		return splitid;
	}
	
	/** remove all noninformative positions and all positions which contain more than one mutation
	 * 
	 * @param aln
	 * @return
	 */
	public static HashMap [] removeNonPolymSites(HashMap aln, HashMap allowedsplits){
		HashMap infpos = new HashMap();
		HashMap newseq = new HashMap();
		for(Iterator it = aln.keySet().iterator(); it.hasNext();){
			String id = (String)it.next();
			newseq.put(id, "");
		}
		int N = ((String)aln.get(aln.keySet().iterator().next())).length();
		for(int i = 0; i < N; i++){
			List split = getPolymorphismSplit(aln, i);
			if(split == null)
				continue;
			String splitid = getSplitId(split);
			if(allowedsplits.get(splitid) != null){
				for(Iterator it = aln.keySet().iterator(); it.hasNext();){
					String id = (String)it.next();
					String sequence = (String)aln.get(id);
					String newsequence = ((String)newseq.get(id))+sequence.charAt(i);
					newseq.put(id, newsequence);
					infpos.put(newsequence.length()-1, i);
				}
			}
		}
		HashMap [] ret = new HashMap[2];
		ret[0] = newseq;
		ret[1] = infpos;
		return ret;
	}
	
	public static HashMap rndPosInAln(HashMap aln, MersenneTwister mt, int size){
		
		int N = ((String)aln.get(aln.keySet().iterator().next())).length();
		char [] dummysequence = new char[N];
		for(int i = 0; i < N; i++)
			dummysequence[i]='A';
		HashMap newseq = new HashMap();
		for(Iterator it = aln.keySet().iterator(); it.hasNext();){
			String id = (String)it.next();
			newseq.put(id, new String(dummysequence));
		}
		
		HashMap allp = PmorphStats.splitFreqs(aln);
		ArrayList allunfolded = new ArrayList();
		for(Iterator it = allp.keySet().iterator(); it.hasNext();){
			String key = (String)it.next();
			String [] split = key.split(";");
			int freq = (Integer)allp.get(key);
			for(int k = 0; k < freq; k++){
				allunfolded.add(split);
			}
		}
				
		IntArrayList positions = new IntArrayList();
		for(int i = 0; i < N; i++)
			positions.add(i);
		
		for(int i = 0; i < size; i++){
			int idx = mt.choose(0,allunfolded.size()-1);
			String [] split = (String [])allunfolded.get(idx);
			allunfolded.remove(idx);
			idx = mt.choose(0,positions.size()-1);
			int pos = positions.get(idx);
			positions.remove(idx);
			for(int j = 0; j < split.length; j++){
				String id = (String)split[j];
				String sequence = (String)newseq.get(id);
				char [] tmp = sequence.toCharArray();
				tmp[pos] = 'T';
				sequence = new String(tmp);
				newseq.put(id,sequence);					
			}
		}
		return newseq;
	}
	
	public static int seqLength(HashMap aln){
		return ((String)aln.get(aln.keySet().iterator().next())).length();
	}
	
	public static HashMap getBinarySNPs(HashMap seq) throws Exception{
		int N = ((String)seq.get(seq.keySet().iterator().next())).length();
		HashMap lines = new HashMap();
		for(Iterator it = seq.keySet().iterator(); it.hasNext();){
			lines.put(it.next(), new IntArrayList());
		}
		for(int i = 0; i < N; i++){
			List split = PmorphUtils.getPolymorphismSplit(seq, i);
			for(Iterator it = seq.keySet().iterator(); it.hasNext();){
				String key = (String)it.next();
				if(split.contains(key)){
					((IntArrayList)lines.get(key)).add(1);
				}else ((IntArrayList)lines.get(key)).add(0);
			}
		}
		return lines;
	}
}

