package uu.binf.utils;

import java.io.*;
import java.util.*;

import cern.jet.random.engine.MersenneTwister;

import edu.cornell.lassp.houle.RngPack.Ranlux;

public class HashMapUtils {
	
	public static void randomize(HashMap map){
		MersenneTwister ran = new MersenneTwister((int)System.currentTimeMillis());
		randomize(map, ran);
	}
	
	public static void randomize(HashMap map, MersenneTwister ran){
		Vector keys = new Vector(map.keySet());
		int N = keys.size();
		for(int i = 0; i < N; i++){
			int j = (int)Math.floor(ran.uniform(i+1, N-1));
			Object tmp = map.get(keys.get(i));
			map.put(keys.get(i), map.get(keys.get(j)));
			map.put(keys.get(j), tmp);
		}
	}
	
	public static void randomize(HashMap map, List groups){
		Ranlux ran = new Ranlux(System.currentTimeMillis());
		for(int x = 0;  x < groups.size(); x++){
		  Vector g = (Vector)groups.get(x);
		  int N = g.size();
		  for(int i = 0; i < N; i++){
			 int j = (int)Math.floor(ran.uniform(i+1, N-1));
			 Object tmp = map.get(g.get(i));
			 map.put(g.get(i), map.get(g.get(j)));
			 map.put(g.get(j), tmp);
		  }
		}		
	}
	
	public static HashMap readHashMap(String file, String sep) throws Exception{
		String l;
		BufferedReader r = new BufferedReader(new FileReader(file));
		HashMap map = new HashMap();
		while((l = r.readLine())!=null){
			if(l.startsWith("#") || !l.contains(sep))
				continue;
			String [] s = l.split(sep);
			map.put(s[0].trim(), s[1].trim());
		}
		r.close();
		return map;
	}
	
	public static HashMap readHashMapMaxCol(String file, String sep, int idx1, int idx2, int idx3) throws Exception{
		String l;
		BufferedReader r = new BufferedReader(new FileReader(file));
		HashMap col3 = new HashMap();
		HashMap map = new HashMap();
		while((l = r.readLine())!=null){
			if(l.startsWith("#") || !l.contains(sep))
				continue;
			String [] s = l.split(sep);
			if(!map.containsKey(s[0])){
				map.put(s[0].trim(), s[1].trim());
				col3.put(s[0].trim(), Double.parseDouble(s[2].trim()));
			}else{
				if(Double.parseDouble(s[2]) > (Double)col3.get(s[0])){
					map.put(s[0].trim(), s[1].trim());
					col3.put(s[0].trim(), Double.parseDouble(s[2].trim()));
				}
			}
		}
		return map;
	}
	
	public static HashMap readHashMap(String file, String sep, int id1, int id2) throws Exception{
		String l;
		BufferedReader r = new BufferedReader(new FileReader(file));
		HashMap map = new HashMap();
		while((l = r.readLine())!=null){
			if(l.startsWith("#"))
				continue;
			String [] s = l.split(sep);
			if(s.length <= id2 || s.length <= id1)
				continue;
			map.put(s[id1].trim(), s[id2].trim());
		}
		return map;
	}
	
	public static HashMap readHashMapLine(String file, String sep, int id1) throws Exception{
		String l;
		BufferedReader r = new BufferedReader(new FileReader(file));
		HashMap map = new HashMap();
		while((l = r.readLine())!=null){
			if(l.startsWith("#"))
				continue;
			String [] s = l.split(sep);
			if(s.length <= id1)
				continue;
			map.put(s[id1], l);
		}
		return map;
	}
	
	public static HashMap readHashMap(String file, String sep, int id1, int id2, int id3, double t, boolean greater) throws Exception{
		String l;
		BufferedReader r = new BufferedReader(new FileReader(file));
		HashMap map = new HashMap();
		while((l = r.readLine())!=null){
			String [] s = l.split(sep);
			String score = s[id3];
			if(greater){
				if(Double.parseDouble(score) > t)
					map.put(s[id1], s[id2]);
			}else {
				if(Double.parseDouble(score) < t)
					map.put(s[id1], s[id2]);
			}
		}
		return map;
	}
	
	public static HashMap fastaToHashMap(String file) throws Exception{
		BufferedReader r = new BufferedReader(new FileReader(file));
		String l = null;
		HashMap map = new HashMap();
		StringBuffer entry = new StringBuffer();
		String key = null;
		String [] split;
		while((l=r.readLine())!=null){
			if(l.startsWith(">")){
				if(entry.length() != 0)		
					map.put(key,entry.toString());
				//split = l.split(" ");
				key = l.replaceAll(">", "");
				entry.delete(0, entry.length());
				/*if(split.length > 1)
					entry.append(split[1].trim());*/
			}else{
				entry.append(l.trim());
			}
		}
		if(entry.length() != 0)		
			map.put(key,entry.toString());
		return map;
	}
	
	public static List multiFastaToHashMap(String file) throws Exception{
		BufferedReader r = new BufferedReader(new FileReader(file));
		String l = null;
		ArrayList maps = new ArrayList();
		String [] split;
		while((l=r.readLine())!=null){
			HashMap map = new HashMap();
			StringBuffer entry = new StringBuffer();
			String key = null;
			do{
				if(l.startsWith("#"))
					continue;
				if(l.startsWith(">")){
					if(entry.length() != 0)		
						map.put(key,entry.toString());
					split = l.split(">");
					key = split[1].trim().split(" ")[0];
					entry.delete(0, entry.length());
				}else{
					entry.append(l.trim());
				}
			}while(!(l=r.readLine()).startsWith("="));
			if(entry.length() != 0)		
				map.put(key,entry.toString());
			maps.add(map);
		}
		return maps;
	}
	
	public static HashMap fastaToHashMapWithSpaceInName(String file) throws Exception{
		BufferedReader r = new BufferedReader(new FileReader(file));
		String l = null;
		HashMap map = new HashMap();
		StringBuffer entry = new StringBuffer();
		String key = null;
		String [] split;
		while((l=r.readLine())!=null){
			if(l.startsWith(">")){
				if(entry.length() != 0)		
					map.put(key,entry.toString());
				key = l.replaceAll(">", "");
				entry.delete(0,entry.length());
			}else{
				entry.append(l.trim());
			}
		}
		if(entry.length() != 0)		
			map.put(key,entry.toString());
		return map;
	}
	
	public static HashMap listToCountMap(List things){
		HashMap counts = new HashMap();
		for(int i = 0; i < things.size(); i++){
			Object th = things.get(i);
			Integer count = (Integer)counts.get(th);
			if(count == null)
				count = 0;
			counts.put(th, count+1);
		}
		return counts;
	}
	
	public static HashMap listToCountMap(List things, HashMap counts){
		for(int i = 0; i < things.size(); i++){
			Object th = things.get(i);
			Integer count = (Integer)counts.get(th);
			if(count == null)
				count = 0;
			counts.put(th, count+1);
		}
		return counts;
	}
	
	public static Object getRndKey(HashMap map, MersenneTwister mt){
		Vector v = new Vector(map.keySet());
		return v.get(mt.choose(0,v.size()-1));
	}
	
	public static void write(String filename, HashMap map) throws Exception{
		BufferedWriter w = new BufferedWriter(new FileWriter(filename));
		for(Iterator it = map.keySet().iterator(); it.hasNext();){
			String id = (String)it.next();
			w.write(id + " " + map.get(id) +"\n");
		}
		w.flush();
		w.close();
	}

}
